

#define ADC_OUT  0          // channel for AC output sample

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

#define NPWM  200         // 200 for this and 800 for PPWM for 50Hz and 20kHz PWM. Both must be integers.
//#define PPWM  665       // 60.062 Hz - choose one or the other.
#define PPWM  799       // 50.002 Hz

      
uint8_t uf,oen,v1low, lv_stopped,int_f;
uint16_t pcount;
uint32_t vpwr;
int sst;
uint16_t l[NPWM+1];
float pwr;
uint8_t dbounce;
double Input, Output, Setpoint;
double errSum, lastErr,error,dErr;
double kp, ki, kd;
float xv[5],yv[5];
float ac_output, ac_setpoint;

void init_vars(void);

uint16_t pcint;

void setup()
  {
  int i;
  float t,u;

  for(t=0.0,i=0; i <= NPWM; i++,t +=   3.14159/ (float) (NPWM+1))
    {   // 1/2 wave sine lookup table, scaled to (16 bits - 1)
    u = 65535.0 * sin(t);
    l[i] =  (int)u;  
    }
 
  pinMode(9, OUTPUT);     // V1 drive
  pinMode(10, OUTPUT);    // V2 drive
  pinMode(8,INPUT_PULLUP);// on/off button 
  // can NOT let arduino system alter any pins from D0 to D7 via calls to pinMode() etc.. It buggers d5 which I need to pull down low-Z upon boot
  DDRD = 0xfb; // and let D2 be input for INT0 
  noInterrupts();

  TCCR1A = _BV(COM1A1) | _BV(COM1B0) | _BV(COM1B1) | _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  OCR1A = 10;
  ICR1 = PPWM;
  TIMSK1 |= (1 << TOIE1);

  EICRA = 3;                // interrupt on falling edge from pin D2
  EIMSK |= bit (INT0);     // enable it  
  
  cbi(PORTD,5);   // disable gate drive output. data direction flag set by above statement.
  oen = 0;      // output enable. = 1 to run inverter, = 0 to stop inverter
  init_vars();
  
  interrupts();             // enable all interrupts
  TIMSK0 = 0;               // except arduino clock

  sbi(ADCSRA,ADPS2) ;   // ADC clock prescale = 16 now. this saves 100us in main loop with little cost to loop stability
  cbi(ADCSRA,ADPS1) ;
  cbi(ADCSRA,ADPS0) ;  

  Serial.begin(115200);
  }

// This simple code can lock onto and remain in sync with any source within
// 50.12 to 49.88 Hz. Pin D4 is pulled high when in sync, low when out of whack
//
// I look for syno at 22 counts into the 400 count sine wave, this delay lines up with the sync interrupt
// when using my circuit. Other circuits will have different delays.
// Lost sync is arbitrary: I choose greater than +/- 4 counts from the required
// set point of 22 counts.
//
ISR (INT0_vect)
{
if  (int_f == 0) return;  // only once per 50 Hz cycle. Noise! much noise...
int_f = 0;

pcint = pcount +  (v1low == 1?200:0);   // get position in 50Hz waveform, 0 - 400

if (pcint < 22)              // servo around 20 counts
  if (pcount < 399) pcount++; // do not overflow sine wave lookup table
if (pcint > 22)
  if (pcount > 0) pcount--;   // and do not underflow.
if (pcint < 18 || pcint > 26)
     // phase lock lost .Tune up some limits as needed.
    cbi(PORTD,4);  
  else
    sbi(PORTD,4); 
}  



ISR(TIMER1_OVF_vect)        
 {    // 20Khz SPWM, code has only 50uS to run, takes about 10uS
  long c;
  
  c = (l[pcount] * vpwr) >> 16;   // scale sin wave by vpwr 32 bit integer calcs.
  if (v1low == 1)
    OCR1B = c;
  else
    OCR1A = c;
  pcount++;
  if(pcount >= NPWM)
    {
    pcount = 0;
    uf=1;
    if (v1low == 1)
      {
       v1low = 0; 
       TCCR1A =  _BV(COM1A1) |  _BV(WGM11);
       sbi(PORTD,7);  // DSO sync pulse
       cbi(PORTD,7); 
       if (int_f == 1)    // no AC sync signal at all, pull lock output LOW
        cbi(PORTD,4);
       int_f = 1; 
      }
    else
      {
       v1low = 1;
       TCCR1A =   _BV(COM1B1) | _BV(WGM11);
      }
    }
}


void init_vars()      // called prior to setting oen = 1, as well as from setup()
  {
  ICR1 = PPWM;
  lv_stopped = 0;  
  sst = 0;      // slow start counter 0 - 251
  v1low = 1;    // half wave output flag, 0 or 1
  pcount = 0;   // counter for sine wave lookup table
  uf = 0;       // set to 1 to enable AC output voltage feedback code to execute
  pwr = 0.0;    // PWM duty cycle analog, 0.0 to 1.0
  vpwr= 0;      // integer equivalent of pwr, 0 to PPWM
  dbounce = 0;    // debounce counter for input switch 
  // set up PID 
  kp = 0.1;   
  ki = 0.0;
  kd = 0.01;
  errSum = 0.0;
  lastErr = 0.0;
  int_f = 0;
  }


void do_pid()
{   
   double timeChange = 0.01;    // always called at 100Hz, so dT will be 10msec..
   error = Setpoint - Input;
   errSum += (error * timeChange);
   dErr = (error - lastErr) / timeChange;
   Output = kp * error + ki * errSum + kd * dErr;
   lastErr = error;
}


void loop()
{
  float ch0;

  ch0 = (float)analogRead(ADC_OUT) / 1024.0; // scale 0 - 5V to 0.0 to 1.0
  // filter AC output 1000 sample rate, Fc = 10, 4 pole Bessel LP
  // thanks to  https://www-users.cs.york.ac.uk/~fisher/mkfilter/trad.html
  xv[0] = xv[1]; xv[1] = xv[2]; xv[2] = xv[3]; xv[3] = xv[4];
  xv[4] = ch0 / 2.259747797e+05;
  yv[0] = yv[1]; yv[1] = yv[2]; yv[2] = yv[3]; yv[3] = yv[4];
  yv[4] =   (xv[0] + xv[4]) + 4 * (xv[1] + xv[3]) + 6 * xv[2]
                     + ( -0.7428578687 * yv[0]) + (  3.1954036268 * yv[1])
                     + ( -5.1599230905 * yv[2]) + (  3.7073065280 * yv[3]);
  ac_output = yv[4];

  if(uf == 1  )    // this will execute at 100Hz
      {            // it runs at near zero volt crossing of AC output       
      uf=0;
  
      ac_setpoint = 0.56;       // for same Vfb as EG8010.
                                // change AC output voltage via Vfb voltage divider trimpot
      if (oen == 1) sst++;   
      if (sst > 251)sst = 251;  
      if (oen == 0) sst--; 
      if (sst <= 0)
        {
          sst = 0;
          cbi(PORTD,5);   // pull IR2184 shutdown LOW to disable gate drive output
        }
      else
        {
        sbi(PORTD,5);     // pull it HIGH, to enable output
        }

      if(sst > 0 && sst < 250)     // slow start timer counts to 250 with gentle ramp up, then switches over to PID control. (gulp!)
        {                           // sst can increase or decrease, giving the slow start and slow stop.
        if (oen == 1)             // slow stop? I think of it as de-gauss
          { 
          if(ac_output < ac_setpoint)
            pwr += 1.0/250.0;     // bang-bang control, should AC output reach setpoint prior to PID
          else
            pwr -= 1.0/250.0;
          }
        else
          pwr -= 1.0 /250.0;
        }
        
      if (sst == 251)       // once sst = 251, change to PID. sst will not be incremented further
        { 
        /*  
        if(ac_output < ac_setpoint)
            pwr += 1.0/250.0;     // bang-bang control, should AC output reach setpoint prior to PID
          else
            pwr -= 1.0/250.0;
        */    
        Setpoint = ac_setpoint;     
        Input = ac_output;  
        do_pid();
        pwr = pwr + Output;
        
        }
        
      if (pwr > 0.99) pwr = 0.99;   // clamp. Do math in floating point to prevent any integer overflow
      if (pwr < 0.01) pwr = 0.01; // it's slower. We could do clamp after conversion to int but it's your mosfets...
      vpwr = (int)((float)PPWM * pwr);     

      check_switch();
      Serial.print(" -50 450 ");
      Serial.println(pcint);
      }
}

void check_switch()
{   // called at 100 Hz
  dbounce++;  
  if (dbounce > 100) dbounce = 100;
  if ((PINB & 1) == 0)     // momentary button pressed, pin 8, pulling internal pullup low
    {
    if (dbounce > 80)     // 0.8 seconds must elapse before button state changes
      {  
      if(oen == 0)
        {
          init_vars();
          oen = 1;    // enable it.
        }
      else
        oen = 0;   // stop inverter
      dbounce=0;
      }       
    }
}


